@@ -30,20 +30,19 @@ class NaNLabelEncoder(BaseEstimator, TransformerMixin):
         self.warn = warn
         super().__init__()
 
-    def fit_transform(self, y: pd.Series) -> np.ndarray:
+    def fit_transform(self, y: pd.Series, overwrite: bool = False) -> np.ndarray:
         """
         Fit and transform data.
 
         Args:
             y (pd.Series): input data
+            overwrite (bool): if to overwrite current mappings or if to add to it.
 
         Returns:
             np.ndarray: encoded data
         """
-        if self.add_nan:
-            self.fit(y)
-            return self.transform(y)
-        return super().transform(y)
+        self.fit(y, overwrite=overwrite)
+        return self.transform(y)
 
     @staticmethod
     def is_numeric(y: pd.Series) -> bool:
@@ -59,26 +58,40 @@ class NaNLabelEncoder(BaseEstimator, TransformerMixin):
         """
         return y.dtype.kind in "bcif" or (isinstance(y, pd.CategoricalDtype) and y.cat.categories.dtype.kind in "bcif")
 
-    def fit(self, y: pd.Series):
+    def fit(self, y: pd.Series, overwrite: bool = False):
         """
         Fit transformer
 
         Args:
             y (pd.Series): input data to fit on
+            overwrite (bool): if to overwrite current mappings or if to add to it.
 
         Returns:
             NaNLabelEncoder: self
         """
+        if not overwrite and hasattr(self, "classes_"):
+            offset = len(self.classes_)
+        else:
+            offset = 0
+            self.classes_ = {}
+
+        # determine new classes
         if self.add_nan:
             if self.is_numeric(y):
                 nan = np.nan
             else:
                 nan = "nan"
-            self.classes_ = {nan: 0}
-            for idx, val in enumerate(np.unique(y)):
-                self.classes_[val] = idx + 1
+            self.classes_[nan] = 0
+            idx = 1
         else:
-            self.classes_ = {val: idx for idx, val in enumerate(np.unique(y))}
+            idx = 0
+
+        idx += offset
+        for val in np.unique(y):
+            if val not in self.classes_:
+                self.classes_[val] = idx
+                idx += 1
+
         self.classes_vector_ = np.array(list(self.classes_.keys()))
         return self
 
@@ -342,12 +355,12 @@ class TorchNormalizer(BaseEstimator, TransformerMixin):
                 q_25 = y.kthvalue(int(len(y) * 0.25), dim=-1).values
             elif isinstance(y, np.ndarray):
                 self.center_ = np.median(y, axis=-1)
-                q_75 = np.percentiley(y, 75, axis=-1)
-                q_25 = np.percentiley(y, 25, axis=-1)
+                q_75 = np.percentile(y, 75, axis=-1)
+                q_25 = np.percentile(y, 25, axis=-1)
             else:
                 self.center_ = np.median(y)
-                q_75 = np.percentiley(y, 75)
-                q_25 = np.percentiley(y, 25)
+                q_75 = np.percentile(y, 75)
+                q_25 = np.percentile(y, 25)
             self.scale_ = (q_75 - q_25) / 2.0 + self.eps
         if not self.center:
             self.scale_ = self.center_
@@ -355,6 +368,14 @@ class TorchNormalizer(BaseEstimator, TransformerMixin):
                 self.center_ = torch.zeros_like(self.center_)
             else:
                 self.center_ = np.zeros_like(self.center_)
+
+        if (np.asarray(self.scale_) < 1e-7).any():
+            warnings.warn(
+                "scale is below 1e-7 - consider not centering "
+                "the data or using data with higher variance for numerical stability",
+                UserWarning,
+            )
+
         return self
 
     def transform(
@@ -514,16 +535,16 @@ class GroupNormalizer(TorchNormalizer):
         if len(self.groups) == 0:
             assert not self.scale_by_group, "No groups are defined, i.e. `scale_by_group=[]`"
             if self.method == "standard":
-                self.norm_ = [np.mean(y), np.std(y) + self.eps]  # center and scale
+                self.norm_ = {"center": np.mean(y), "scale": np.std(y) + self.eps}  # center and scale
             else:
                 quantiles = np.quantile(y, [0.25, 0.5, 0.75])
-                self.norm_ = [
-                    quantiles[1],
-                    (quantiles[2] - quantiles[0]) / 2.0 + self.eps,
-                ]  # center and scale
+                self.norm_ = {
+                    "center": quantiles[1],
+                    "scale": (quantiles[2] - quantiles[0]) / 2.0 + self.eps,
+                }  # center and scale
             if not self.center:
-                self.norm_[1] = self.norm_[0]
-                self.norm_[0] = 0.0
+                self.norm_["scale"] = self.norm_["center"] + self.eps
+                self.norm_["center"] = 0.0
 
         elif self.scale_by_group:
             if self.method == "standard":
@@ -552,7 +573,7 @@ class GroupNormalizer(TorchNormalizer):
             if not self.center:  # swap center and scale
 
                 def swap_parameters(norm):
-                    norm["scale"] = norm["center"]
+                    norm["scale"] = norm["center"] + self.eps
                     norm["center"] = 0.0
                     return norm
 
@@ -581,9 +602,25 @@ class GroupNormalizer(TorchNormalizer):
                     )[["center", "scale"]]
                 )
             if not self.center:  # swap center and scale
-                self.norm_["scale"] = self.norm_["center"]
+                self.norm_["scale"] = self.norm_["center"] + self.eps
                 self.norm_["center"] = 0.0
             self.missing_ = self.norm_.median().to_dict()
+
+        if (
+            (self.scale_by_group and any([(self.norm_[group]["scale"] < 1e-7).any() for group in self.groups]))
+            or (not self.scale_by_group and isinstance(self.norm_["scale"], float) and self.norm_["scale"] < 1e-7)
+            or (
+                not self.scale_by_group
+                and not isinstance(self.norm_["scale"], float)
+                and (self.norm_["scale"] < 1e-7).any()
+            )
+        ):
+            warnings.warn(
+                "scale is below 1e-7 - consider not centering "
+                "the data or using data with higher variance for numerical stability",
+                UserWarning,
+            )
+
         return self
 
     @property
@@ -664,7 +701,7 @@ class GroupNormalizer(TorchNormalizer):
         assert len(group_names) == len(self.groups), "Passed groups and fitted do not match"
 
         if len(self.groups) == 0:
-            params = np.asarray(self.norm_).squeeze()
+            params = np.array([self.norm_["center"], self.norm_["scale"]])
         elif self.scale_by_group:
             norm = np.array([1.0, 1.0])
             for group, group_name in zip(groups, group_names):
@@ -692,7 +729,7 @@ class GroupNormalizer(TorchNormalizer):
             pd.DataFrame: dataframe with scaling parameterswhere each row corresponds to the input dataframe
         """
         if len(self.groups) == 0:
-            norm = np.asarray(self.norm_).reshape(1, -1)
+            norm = np.asarray([self.norm_["center"], self.norm_["scale"]]).reshape(1, -1)
         elif self.scale_by_group:
             norm = [
                 np.prod(
