@@ -1,14 +1,13 @@
 """
 The temporal fusion transformer is a powerful predictive model for forecasting timeseries
 """
-from copy import deepcopy
-from typing import Callable, Dict, List, Tuple, Union
+from typing import Dict, List, Tuple, Union
 
 from matplotlib import pyplot as plt
 import numpy as np
-from pytorch_lightning.metrics import Metric as LightningMetric
 import torch
 from torch import nn
+from torchmetrics import Metric as LightningMetric
 
 from pytorch_forecasting.data import TimeSeriesDataSet
 from pytorch_forecasting.data.encoders import NaNLabelEncoder
@@ -23,7 +22,7 @@ from pytorch_forecasting.models.temporal_fusion_transformer.sub_modules import (
     InterpretableMultiHeadAttention,
     VariableSelectionNetwork,
 )
-from pytorch_forecasting.utils import autocorrelation, create_mask, integer_histogram, padded_stack, to_list
+from pytorch_forecasting.utils import autocorrelation, create_mask, detach, integer_histogram, padded_stack, to_list
 
 
 class TemporalFusionTransformer(BaseModelWithCovariates):
@@ -33,7 +32,6 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
         lstm_layers: int = 1,
         dropout: float = 0.1,
         output_size: Union[int, List[int]] = 7,
-        n_targets: int = 1,
         loss: MultiHorizonMetric = None,
         attention_head_size: int = 4,
         max_encoder_length: int = 10,
@@ -91,7 +89,6 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
             dropout: dropout rate
             output_size: number of outputs (e.g. number of quantiles for QuantileLoss and one target or list
                 of output sizes).
-            n_targets: number of targets. Defaults to 1.
             loss: loss function taking prediction and targets
             attention_head_size: number of attention heads (4 is a good default)
             max_encoder_length: length to encode (can be far longer than the decoder length but does not have to be)
@@ -318,7 +315,7 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
         # output processing -> no dropout at this late stage
         self.pre_output_gate_norm = GateAddNorm(self.hparams.hidden_size, dropout=None, trainable_add=False)
 
-        if self.hparams.n_targets > 1:  # if to run with multiple targets
+        if self.n_targets > 1:  # if to run with multiple targets
             self.output_layer = nn.ModuleList(
                 [nn.Linear(self.hparams.hidden_size, output_size) for output_size in self.hparams.output_size]
             )
@@ -345,29 +342,7 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
         """
         # add maximum encoder length
         new_kwargs = dict(max_encoder_length=dataset.max_encoder_length)
-
-        # infer output size
-        def get_output_size(normalizer, loss):
-            if isinstance(loss, QuantileLoss):
-                return len(loss.quantiles)
-            elif isinstance(normalizer, NaNLabelEncoder):
-                return len(normalizer.classes_)
-            else:
-                return 1
-
-        loss = kwargs.get("loss", QuantileLoss())
-        # handle multiple targets
-        new_kwargs["n_targets"] = len(dataset.target_names)
-        if new_kwargs["n_targets"] > 1:  # try to infer number of ouput sizes
-            new_kwargs["output_size"] = [
-                get_output_size(normalizer, l)
-                for normalizer, l in zip(dataset.target_normalizer.normalizers, loss.metrics)
-            ]
-            if not isinstance(loss, MultiLoss):
-                loss = MultiLoss([deepcopy(loss)] * new_kwargs["n_targets"])
-            new_kwargs["loss"] = loss
-        else:
-            new_kwargs["output_size"] = get_output_size(dataset.target_normalizer, loss)
+        new_kwargs.update(cls.deduce_default_output_parameters(dataset, kwargs, QuantileLoss()))
 
         # update defaults
         new_kwargs.update(kwargs)
@@ -517,53 +492,39 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
         # skip connection over temporal fusion decoder (not LSTM decoder despite the LSTM output contains
         # a skip from the variable selection network)
         output = self.pre_output_gate_norm(output, lstm_output[:, max_encoder_length:])
-        if self.hparams.n_targets > 1:  # if to use multi-target architecture
+        if self.n_targets > 1:  # if to use multi-target architecture
             output = [output_layer(output) for output_layer in self.output_layer]
         else:
             output = self.output_layer(output)
 
-        return dict(
-            prediction=output,
+        return self.to_network_output(
+            prediction=self.transform_output(output, target_scale=x["target_scale"]),
             attention=attn_output_weights,
             static_variables=static_variable_selection,
             encoder_variables=encoder_sparse_weights,
             decoder_variables=decoder_sparse_weights,
             decoder_lengths=decoder_lengths,
             encoder_lengths=encoder_lengths,
-            groups=x["groups"],
-            decoder_time_idx=x["decoder_time_idx"],
-            target_scale=x["target_scale"],
         )
 
     def on_fit_end(self):
         if self.log_interval > 0:
             self.log_embeddings()
 
-    def step(self, x, y, batch_idx):
-        """
-        run at each step for training or validation
-        """
-        # extract data and run model
-        log, out = super().step(x, y, batch_idx)
-        # calculate interpretations etc for latter logging
+    def create_log(self, x, y, out, batch_idx, **kwargs):
+        log = super().create_log(x, y, out, batch_idx, **kwargs)
         if self.log_interval > 0:
+            log["interpretation"] = self._log_interpretation(out)
+        return log
 
-            def detach(v):
-                if isinstance(v, torch.Tensor):
-                    return v.detach()
-                elif isinstance(v, (list, tuple)) and len(v) > 0 and isinstance(v[0], torch.Tensor):
-                    return [vp.detach() for vp in v]
-                else:
-                    return v
-
-            detached_output = {name: detach(out_part) for name, out_part in out.items()}
-            interpretation = self.interpret_output(
-                detached_output,
-                reduction="sum",
-                attention_prediction_horizon=0,  # attention only for first prediction horizon
-            )
-            log["interpretation"] = interpretation
-        return log, out
+    def _log_interpretation(self, out):
+        # calculate interpretations etc for latter logging
+        interpretation = self.interpret_output(
+            detach(out),
+            reduction="sum",
+            attention_prediction_horizon=0,  # attention only for first prediction horizon
+        )
+        return interpretation
 
     def epoch_end(self, outputs):
         """
@@ -683,6 +644,7 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
         add_loss_to_title: bool = False,
         show_future_observed: bool = True,
         ax=None,
+        **kwargs,
     ) -> plt.Figure:
         """
         Plot actuals vs prediction and attention
@@ -702,7 +664,13 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
 
         # plot prediction as normal
         fig = super().plot_prediction(
-            x, out, idx=idx, add_loss_to_title=add_loss_to_title, show_future_observed=show_future_observed, ax=ax
+            x,
+            out,
+            idx=idx,
+            add_loss_to_title=add_loss_to_title,
+            show_future_observed=show_future_observed,
+            ax=ax,
+            **kwargs,
         )
 
         # add attention on secondary axis
@@ -739,7 +707,7 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
 
         # attention
         fig, ax = plt.subplots()
-        attention = interpretation["attention"].cpu()
+        attention = interpretation["attention"].detach().cpu()
         attention = attention / attention.sum(-1).unsqueeze(-1)
         ax.plot(
             np.arange(-self.hparams.max_encoder_length, attention.size(0) - self.hparams.max_encoder_length), attention
@@ -761,13 +729,13 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
             return fig
 
         figs["static_variables"] = make_selection_plot(
-            "Static variables importance", interpretation["static_variables"].cpu(), self.static_variables
+            "Static variables importance", interpretation["static_variables"].detach().cpu(), self.static_variables
         )
         figs["encoder_variables"] = make_selection_plot(
-            "Encoder variables importance", interpretation["encoder_variables"].cpu(), self.encoder_variables
+            "Encoder variables importance", interpretation["encoder_variables"].detach().cpu(), self.encoder_variables
         )
         figs["decoder_variables"] = make_selection_plot(
-            "Decoder variables importance", interpretation["decoder_variables"].cpu(), self.decoder_variables
+            "Decoder variables importance", interpretation["decoder_variables"].detach().cpu(), self.decoder_variables
         )
 
         return figs
@@ -779,7 +747,7 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
         # extract interpretations
         interpretation = {
             # use padded_stack because decoder length histogram can be of different length
-            name: padded_stack([x["interpretation"][name] for x in outputs], side="right", value=0).sum(0)
+            name: padded_stack([x["interpretation"][name].detach() for x in outputs], side="right", value=0).sum(0)
             for name in outputs[0]["interpretation"].keys()
         }
         # normalize attention with length histogram squared to account for: 1. zeros in attention and
@@ -811,7 +779,12 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
         # log lengths of encoder/decoder
         for type in ["encoder", "decoder"]:
             fig, ax = plt.subplots()
-            lengths = padded_stack([out["interpretation"][f"{type}_length_histogram"] for out in outputs]).sum(0).cpu()
+            lengths = (
+                padded_stack([out["interpretation"][f"{type}_length_histogram"] for out in outputs])
+                .sum(0)
+                .detach()
+                .cpu()
+            )
             if type == "decoder":
                 start = 1
             else:
@@ -832,5 +805,5 @@ class TemporalFusionTransformer(BaseModelWithCovariates):
         for name, emb in self.input_embeddings.items():
             labels = self.hparams.embedding_labels[name]
             self.logger.experiment.add_embedding(
-                emb.weight.data.cpu(), metadata=labels, tag=name, global_step=self.global_step
+                emb.weight.data.detach().cpu(), metadata=labels, tag=name, global_step=self.global_step
             )
